/**
  Sample code for thread migration
  in package NET.shudo.moba.

  Copy this file to the directory which is in CLASSPATH and compile.
    % javac MigrationRunnable.java
  Make sure that this class (MigrationRunnable) is found in your CLASSPATH
  at the source host.
  Type the following command at the source and destination host
    % (somewhere)/moba/bin/mobaplace &
  to invoke MOBA place which receive threads from the network.
  Now execute at the source host.
    % java MigrationRunnable <destination_host_name>

  @author SHUDOH Kazuyuki
 */

import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.PrintStream;
import java.io.IOException;

import NET.shudo.moba.lang.MobaThread;
import NET.shudo.moba.io.ObjectWriter;
import NET.shudo.moba.io.ThreadWriter;


class MigrationRunnable {
  private void start(String[] argv) {
    // åɤεư
    Runnable r;
    if (argv.length > 0)
      r = new ARunnableMigrant(argv[0]);
    else
      r = new ARunnableMigrant();
    Thread t = new MobaThread(r);	// ¸ǽ Thread 饹
    t.start();

    // ưåɤνλԤ
    boolean joined = true;
    do {
      try {
	t.join();
      } catch (InterruptedException e) { joined = false; }
    } while (!joined);
  }

  public static void main(String[] argv) {
    new MigrationRunnable().start(argv);
  }
}


/**
 * ư륹å
 */
class ARunnableMigrant implements Runnable {
  private static final String DEFAULT_HOST = "host.defaultdomain.com";
  private static final String THREAD_FNAME = "output";

  private String host = null;

  public ARunnableMigrant() {
    this(DEFAULT_HOST);
  }
  public ARunnableMigrant(String host) {
    this.host = host;
  }


  public void run() {
    System.out.println("Now externalize myself(Thread#currentThread()).");

/* for debugging
    try {
      ObjectWriter out = new ThreadWriter(new FileOutputStream(THREAD_FNAME));
      out.writeObject(Thread.currentThread());
      out.close();
    }
    catch (IOException e) { e.printStackTrace(); System.exit(1); }
*/

    // ư or fork
    try {
      // migration
      MobaThread.goTo(this.host);
      // fork
//      MobaThread.fork(this.host);
    }
    catch (IOException e) { e.printStackTrace(); System.exit(1); }


    if (((MobaThread)Thread.currentThread()).isInternalized()) {
      System.out.println("I've migrated !");
//      System.out.flush();

/* for debugging
      try {
	PrintStream out = new PrintStream(new FileOutputStream("migrated"));
	out.println("I've migrated !");
	out.close();
      }
      catch (IOException e) { e.printStackTrace(); System.exit(1); }
*/
    }
    else {
      System.out.println("Migration is done.");
	// This message is never shown
	// because he, current thread is gone. :)
    }
  }
}
