/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.io;

import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import NET.shudo.moba.reflect.*;
import NET.shudo.moba.lang.GlobalReference;	// for method loadLibrary()


/**
 * Class externalizer<BR>
 * VM Υ饹򥯥饹եΥեޥåȤǽϤ롣
 *
 * @author SHUDOH Kazuyuki
 */
public final
class ClassWriter extends DataOutputStream {
  public ClassWriter(OutputStream out) {
    super(out);

    // native methods 饤֥Υ
    GlobalReference.loadLibrary();
  }

  public ClassWriter() {
    this(System.out);
  }


  public void writeClass(Class clazz) throws IOException {
    // attrib. ̾ index
    int utfConstantValueIndex = ConstantPool.utf8Index(clazz, "ConstantValue");
    int utfCodeIndex = ConstantPool.utf8Index(clazz, "Code");
    int utfExceptionsIndex = ConstantPool.utf8Index(clazz, "Exceptions");
    int utfLineNoTblIndex = ConstantPool.utf8Index(clazz, "LineNumberTable");
    int utfLVarTblIndex = ConstantPool.utf8Index(clazz, "LocalVariableTable");
    int utfSourceFileIndex = ConstantPool.utf8Index(clazz, "SourceFile");
    int utfDeprecatedIndex = ConstantPool.utf8Index(clazz, "Deprecated");


    MetaClass mc = MetaClass.get(clazz);
    int i, j;

    // 饹γƼ
    int cpCount = mc.constantpoolCount();
    int intfCount = mc.interfaceCount();
    int fieldCount = mc.fieldCount();
    int methodCount = mc.methodCount();
    String sourceName = mc.sourceName();

    // 줾 attrib. ¸ߤ뤫ʤ (ѿ *Exists)
    boolean sourceFileExists =
	((sourceName != null) && (utfSourceFileIndex > 0));


    // magic number
    writeInt(MetaClass.magicNumber());	// 0xcafebabe

    // major, minor version
    writeShort(mc.minorVersion());
    writeShort(mc.majorVersion());

    // constantpool count
    writeShort(cpCount);

    // constantpool
    ConstantPool.writeInClassfile(this, clazz);

    // access flag
    writeShort(Access.maskInClassfile(mc.access()));

    // this class
    writeShort(ConstantPool.classIndex(clazz, clazz));

    // super class
    {
      MetaClass superMc = mc.superClass();
      if (superMc != null) {
	writeShort(ConstantPool.classIndex(clazz,
			mc.superClass().classClass()));
      }
      else {
	writeShort(0);
      }
    }

    // interface count
    writeShort(intfCount);

    // interface
    mc.writeIFInClassfile(this);

    // field count
    writeShort(fieldCount);


    // field
    for (i = 0; i < fieldCount; i++) {
      MetaField mf = null;
      int access = 0;
      try {
	mf = mc.field(i);
      }
      catch (FieldNotFoundException e) {
	// not reached.
      }
      access = mf.access();

      // access
      writeShort(Access.maskInClassfile(access));
      // name index
      writeShort(ConstantPool.utf8Index(clazz, mf.name()));
      // signature index
      writeShort(ConstantPool.utf8Index(clazz, mf.signature()));
      // attrib. count, attrib.
      if (Access.isStatic(access) && Access.isValknown(access)) {
	if (utfConstantValueIndex >= 0) {
	  // attrib. count
	  writeShort(1);

	  // attrib. `ConstantValue'
	  writeShort(utfConstantValueIndex);	// `ConstantValue' in CP
	  writeInt(2);		// attrib. length
	  if (mf.isDouble()) {
	    writeShort(ConstantPool.doubleIndex(clazz,
					mf.getDoubleValue(null)));
	  }
	  else if (mf.isLong()) {
	    writeShort(ConstantPool.longIndex(clazz, mf.getLongValue(null)));
	  }
	  else if (mf.isFloat()) {
	    writeShort(ConstantPool.floatIndex(clazz, mf.getFloatValue(null)));
	  }
	  else if (mf.isByte() || mf.isChar() || mf.isShort() || mf.isInt()) {
	    writeShort(ConstantPool.intIndex(clazz, mf.getIntValue(null)));
	  }
	  else if (mf.isBoolean()) {
	    writeShort(ConstantPool.booleanIndex(clazz,
					mf.getBooleanValue(null)));
	  }
	  else if (mf.isObject()) {
	    Object obj = mf.getObjectValue(null);
	    if (obj instanceof java.lang.String) {
	      writeShort(ConstantPool.stringIndex(clazz, (String)obj));
	    }
	    else {
	      // not reached.
	      writeShort(0);
	    }
	  }
	  else {
	    // not reached.
	    writeShort(0);
	  }	// եɤηȽ
	}
	else
	  writeShort(0);
		// CP  (UTF8) `ConstantValue' Ĥ뤫
      }
      else
	writeShort(0);
		// (mod & ACC_STATIC) && (mod & ACC_VALKNOWN)
    }	// եɤé


    // method count
    writeShort(methodCount);


    // method
    for (i = 0; i < methodCount; i++) {
      MetaMethod mm = null;
      int access = 0;
      try {
	mm = mc.method(i);
      }
      catch (FieldNotFoundException e) {
	// not reached.
      }
      access = mm.access();

      // access
      writeShort(Access.maskInClassfile(access));
      // name index
      writeShort(ConstantPool.utf8Index(clazz, mm.name()));
      // signature index
      writeShort(ConstantPool.utf8Index(clazz, mm.signature()));
      // attrib. count, attrib.
      //	attribute `Deprecated' б
      {
	int methodAttrCount = 0;	// attrib. Code
	int thrownExLength = mm.thrownExceptionLength();
	boolean thrownExExists =
		((thrownExLength > 0) && (utfExceptionsIndex > 0));

	// attrib. count
	if (!Access.isNative(access))  methodAttrCount++;
	if (thrownExExists)  methodAttrCount++;
	writeShort(methodAttrCount);

	if (!Access.isNative(access)) {
	  // attrib. `Code'
	  {
	    writeShort(utfCodeIndex);	// `Code' in CP

	    long codeAttrLen;
	    long codeLength = mm.codeLength();
	    int catchFrameLength = mm.catchFrameLength();
	    int lineNoTableLength = mm.lineNoTableLength();
	    int localVarTableLength = mm.localVarTableLength();
	    boolean lineNoTableExists =
		((lineNoTableLength > 0) && (utfLineNoTblIndex > 0));
	    boolean localVarTableExists =
		((localVarTableLength > 0) && (utfLVarTblIndex > 0));

	    // attrib. length
	    codeAttrLen = 12 + codeLength + 8 * catchFrameLength;
	    if (lineNoTableExists) {
	      codeAttrLen += 8 + 4 * lineNoTableLength;
	    }
	    if (localVarTableExists) {
	      codeAttrLen += 8 + 10 * localVarTableLength;
	    }
	    writeInt((int)codeAttrLen);	// down cast

	    // max stack
	    writeShort(mm.maxStack());
	    // max locals
	    writeShort(mm.maxLocals());
	    // code length
	    writeInt((int)codeLength);	// down cast

	    // code
	    mm.writeCode(this);

	    // catch frame length
	    writeShort(catchFrameLength);
	    for (j = 0; j < catchFrameLength; j++) {
	      writeShort(mm.catchFrameStartPC(j));
	      writeShort(mm.catchFrameEndPC(j));
	      writeShort(mm.catchFrameHandlerPC(j));
	      writeShort(mm.catchFrameExceptionIndexInCP(j));
	    }

	    // attrib. count
	    int codeAttrCount = 0;
	    if (lineNoTableExists)  codeAttrCount++;
	    if (localVarTableExists)  codeAttrCount++;
	    writeShort(codeAttrCount);

	    // attrib. `LineNumberTable'
	    if (lineNoTableExists) {
	      writeShort(utfLineNoTblIndex);	// `LineNumberTable' in CP

	      // attrib. length
	      int attrLen = 2 + 4 * lineNoTableLength;
	      writeInt(attrLen);

	      writeShort(lineNoTableLength);

	      for (j = 0; j < lineNoTableLength; j++) {
		writeShort(mm.lineNoTablePC(j));
		writeShort(mm.lineNoTableLineNo(j));
	      }
	    }

	    // attrib. `LocalVariableTable'
	    if (localVarTableExists) {
	      writeShort(utfLVarTblIndex);	// `LovalVariableTable' in CP

	      // attrib .length
	      int attrLen = 2 + 10 * localVarTableLength;
	      writeInt(attrLen);

	      writeShort(localVarTableLength);

	      for (j = 0; j < localVarTableLength; j++) {
		writeShort(mm.localVarTablePC(j));
		writeShort(mm.localVarTableLife(j));
		writeShort(mm.localVarTableNameIndex(j));
		writeShort(mm.localVarTableSigIndex(j));
		writeShort(mm.localVarTableSlot(j));
	      }
	    }
	  }	// attrib. `Code'
	}	// native method ɤ

	// attrib. `Exceptions'
	if (thrownExExists) {
	  writeShort(utfExceptionsIndex);	// `Exceptions' in CP

	  // attrib. length
	  int attrLen = 2 + 2 * thrownExLength;
	  writeInt(attrLen);

	  writeShort(thrownExLength);

	  for (j = 0; j < thrownExLength; j++) {
	    writeShort(mm.thrownExceptionIndexInCP(j));
	  }
	}
      }	// attrib. count, attrib.
    }	// ᥽åɤé


    // attrib. count
    int classAttribCount = 0;
    if (sourceFileExists)  classAttribCount++;
    writeShort(classAttribCount);

    // attrib. `Source'
    if (sourceFileExists) {
      writeShort(utfSourceFileIndex);	// `Code' in CP

      // attrib. length
      writeInt(2);

      writeShort(ConstantPool.utf8Index(clazz, sourceName));
    }


    this.flush();
  }
}
