/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.io;

import java.io.InputStream;
import java.io.DataInputStream;
import java.io.BufferedInputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Stack;
import NET.shudo.moba.reflect.*;


/**
 * Object de-serializer
 * java.io.ObjectInputStream롣
 *
 * @author SHUDOH Kazuyuki
 */
public
class ObjectReader extends DataInputStream {
  /**
   * Ǥ de-serialize  object ɽ
   */
  private Vector objectArray = new Vector();

  private int nullCount = 0;

  /**
   * 饹
   */
  private ClassLoader loader = null;
  /**
   * Ȥ٤饹֤
   */
  public ClassLoader classLoader() { return this.loader; }
  /**
   * Ȥ٤饹ꤹ롣
   */
  public void classLoader(ClassLoader loader) { this.loader = loader; }


  public ObjectReader(InputStream in, ClassLoader loader) throws IOException {
    super((in.markSupported()) ? in : (new BufferedInputStream(in)));

    if (!this.markSupported()) {
      // not reached.
      throw new IOException("failed to make markSupported stream");
    }

    this.loader = loader;
  }
  public ObjectReader(InputStream in) throws IOException {
    this(in, null);
  }
  public ObjectReader() throws IOException {
    this(System.in, null);
  }


  /**
   * subclassΥ饹Ф򤹤뤿hook
   * 󥹥󥹤֤
   *
   * @return true if the specified class is the target class.
   */
  protected Object newInstanceOfSpecificClass(MetaClass mc) {
//	throws InstantiationException, IllegalAccessException {
    return mc.newInstance();
  }

  /**
   * subclassΥ饹Ф򤹤뤿hook
   * ꥯ饹ɤȽꤹ롣
   *
   * @return true if the specified class is the target class.
   */
  protected boolean isSpecificClass(Class clazz) {
    return false;
  }

  /**
   * subclassΥ饹Ф򤹤뤿hook<BR>
   * 󥹥󥹤ΥФꤷ֤<BR>
   * Υ饹Ȥ supeclass ˤĤƤǤ롣
   *
   * @param obj instantiate ѤߤΥ֥
   * @return true if the specified class is the target class.
   */
  protected boolean readSpecificClass(Class clazz, Object obj)
	throws StreamFormatException
		,IOException, ClassNotFoundException
		,InstantiationException, IllegalAccessException {
    return false;
  }

  /**
   * subclassΥ饹Ф򤹤뤿hook<BR>
   * 󥹥󥹤ФԤ
   *
   * @return true if the specified class is the target class.
   */
  protected boolean postReadSpecificClass(Object obj) {
    return false;
  }


  public Object readObject()
	throws StreamFormatException
		,IOException, ClassNotFoundException
		,InstantiationException, IllegalAccessException {
    Object obj = null;
    MetaClass metaClass = null;

    Stack metaClassStack = new Stack();
	/* instance var. ˤƤϤʤ */


    // null reference
    if (readNull()) {
      return null;
    }

    // ǥå
    if ((obj = readIndex()) != null) {
      return obj;
    }

    // ̰륯饹
    obj = checkSpecialClasses();
    if (obj != null) {
      objectArray.addElement(obj);

      return obj;
    }


    // 饹̾ɤ߹
    String className = readUTF();

    metaClass = MetaClass.get(className, this.loader);

    if (metaClass.isArray()) {		// 
      int len = readInt();	// Ĺ
      if (len == -1)  return null;

      obj = (Object)metaClass.newInstance(len);	// 󥪥֥Ȥ

      objectArray.addElement(obj);	// Ͽ

      if (className.startsWith("[L") || className.startsWith("[[")) {
	// ֥ or  
	fillObjectArray(obj, metaClass);
      }
      else {
	// ܷ
	fillBaseTypeArray(obj, className);
      }

      return obj;
    }


    // 󥹥󥹤
    obj = newInstanceOfSpecificClass(metaClass);
    // ä֥ȤϿ
    objectArray.addElement(obj);

    // 饹 readExternal() ư
    if (metaClass.isStreamable()) {
      ((Streamable)obj).readFromStream(this);
      return obj;
    }

    // MetaClass 򥹥åѤ
    while (metaClass != null) {
      metaClassStack.push(metaClass);
      if (isSpecificClass(metaClass.classClass()))
	break;
      metaClass = metaClass.superClass();
    }

    // java.lang.Object  subclass é
    while (!(metaClassStack.empty())) {
      metaClass = (MetaClass)metaClassStack.pop();

      // ObjectReader  subclass ̰륯饹
      if (readSpecificClass(metaClass.classClass(), obj))
	continue;

      // 饹 readObject() ư
      if (metaClass.hasOwnReadObject()) {
	boolean retInvoke;
	retInvoke = invokeOwnObjectReader(metaClass.classClass(), obj);

	int term = readByte();
	if (((byte)term) != (byte)0xff) {
	  throw new StreamFormatException("Illegal stream format");
	}

	continue;
      }

      // Фɤ߼
      for (int i = 0; i < metaClass.fieldCount(); i++) {	// field é
	MetaField field = null;
	try {
	  field = metaClass.field(i);
	}
	catch (FieldNotFoundException e) {
	  // not reached
	  e.printStackTrace();  System.exit(1);
	}

	// serialize ٤ǤϤʤե
	if (!field.isAccessible())  continue;

	// ֥ or 
	else if (field.isObject()) {
	  field.setObjectValue(obj, readObject());
	}
	// ܷ
	else {
	  readBaseType(obj, field);
	}
      } // for (int i = 0; i < metaClass.fieldCount(); i++)  // field é
    }	// while (!(metaClassStack.empty()))  subclass é


    // terminater
    {
      int term = readByte();
      if ((byte)term != (byte)0xff) {
	throw new StreamFormatException(
		"Illegal stream format(can't find terminator)");
      }
    }


    // ְ¸롣
    if (reviseLocDep) {
      // java.io.FileDescriptor#fd
      if (obj instanceof java.io.FileDescriptor) {
//System.out.println("readObject(): FileDescriptor is found.");
//System.out.flush();
	MetaField fdField = null;
	try {
	  fdField = metaClass.field("fd");
	}
	catch (FieldNotFoundException e) {
	  // not reached.
	  System.err.println("a field java.io.FileDescriptor#fd is not found.");
	  e.printStackTrace();
	}
	if (fdField != null) {
	  fdField.setIntValue(obj, -1);	// fd ̵
//System.out.println("readObject(): set FileDescriptor#fd to -1");
//System.out.flush();
	}
      }
    }


    // ̰륯饹ˤĤƤθ
    postReadSpecificClass(obj);


    return obj;
  }


  /**
   * ְ¸뤫ɤΥե饰
   */
  private boolean reviseLocDep = true;

  /**
   * ְ¸뤫ɤꤹ롣
   */
  public boolean reviseLocationDependence(boolean flag) {
    boolean old = reviseLocDep;
    reviseLocDep = flag;
    return old;
  }


  /**
   * 饹 readObject() Ƥ֡
   */
  private native boolean invokeOwnObjectReader(Class classclass, Object obj)
	throws IOException;


  /**
   * demarshal Ѥߥ֥ȤΥơ֥Ȥɽ롣(for debug)
   */
  public void showObjectTable(PrintStream out) {
    int i = 0;
    for (Enumeration e = objectArray.elements(); e.hasMoreElements(); ) {
      out.println(i++ + ": " + (e.nextElement()));
    }
  }


  /**
   *  writeObject()
   */
  private void writeObject(ObjectWriter out) throws IOException {
  }
  /**
   *  readObject()
   */
  private void readObject(ObjectReader in)
	throws IOException, ClassNotFoundException {
  }


  /*
   * 
   */

  private boolean readNull() throws IOException {
    byte b1, b2;

    // null Ϣ³
    if (nullCount > 0) { nullCount--;  return true; }

    mark(2);

    b1 = readByte();  b2 = readByte();
    if (b1 == (byte)0xff) {
      if (b2 == (byte)0xff) {
	return true;
      }
      else if (b2 == (byte)0xfe) {
	nullCount = readShort();
	nullCount--;
	return true;
      }
    }

    reset();
    return false;
  }

  private Object readIndex() throws IOException {
    byte b1, b2;

    mark(2);
    b1 = readByte();  b2 = readByte();
    if ((b1 == (byte)0xff) && (b2 == (byte)0xfd)) {
      int index = readInt();
//System.out.print("registered object[" + index + "]: ");  System.out.flush();
//System.out.println(objectArray.elementAt(index));  System.out.flush();

      return objectArray.elementAt(index);
    }

    reset();
    return null;
  }

  /**
   * ̰륯饹
   */
  private Object checkSpecialClasses() throws IOException {
    Object obj = null;

    // java.lang.String
    if ((obj = readString()) != null) {
      return obj;
    }
    // java.lang.Class
    else if ((obj = readClass()) != null) {
      return obj;
    }

    return null;
  }

  private String readString() throws IOException {
    byte b1, b2;

    mark(2);
    b1 = readByte();  b2 = readByte();
    if ((b1 == (byte)0xff) && (b2 == (byte)0xfc)) {
      String str = readUTF();
      str = str.intern();	// ǰΤ
      return str;
    }

    reset();
    return null;
  }

  private Class readClass() throws IOException {
    byte b1, b2;

    mark(2);
    b1 = readByte();  b2 = readByte();
    if ((b1 == (byte)0xff) && (b2 == (byte)0xfb)) {
      String className;
      Class clazz;

      className = readUTF();
      try {
	clazz = Class.forName(className);
      }
      catch (ClassNotFoundException e) {
	throw new IOException("Class.forName(" + className + ") failed.");
      }
      return clazz;
    }

    reset();
    return null;
  }


  /**
   * ܷɤ߹
   */
  private final void readBaseType(Object obj, MetaField field)
	throws IOException {
    String sig = field.signature();

    if (Signature.isByte(sig)) {
      field.setByteValue(obj, readByte());
    }
    else if (Signature.isChar(sig)) {
      field.setCharValue(obj, readChar());
    }
    else if (Signature.isDouble(sig)) {
      field.setDoubleValue(obj, readDouble());
    }
    else if (Signature.isFloat(sig)) {
      field.setFloatValue(obj, readFloat());
    }
    else if (Signature.isInt(sig)) {
      field.setIntValue(obj, readInt());
    }
    else if (Signature.isLong(sig)) {
      field.setLongValue(obj, readLong());
    }
    else if (Signature.isShort(sig)) {
      field.setShortValue(obj, readShort());
    }
    else if (Signature.isBoolean(sig)) {
      field.setBooleanValue(obj, readBoolean());
    }
    else {
      // not reached.
      // 顼  not implemented yet.
    }
  }

  /**
   * ֥Ȥɤ߹
   */
  private final void fillObjectArray(Object obj, MetaClass mc)
	throws StreamFormatException
		,IOException, ClassNotFoundException
		,InstantiationException, IllegalAccessException {
    if (obj == null)  return;

    Object[] objArray = (Object[]) obj;

    for (int j = 0; j < objArray.length; j++) {
      objArray[j] = readObject();
    }
  }

  /**
   * ܷɤ߹
   */
  private final native void fillBaseTypeArray(Object array, String sig)
	throws IOException;
}
