/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.io;

import java.io.InputStream;
import java.io.IOException;
import NET.shudo.moba.lang.MobaThread;
import NET.shudo.moba.lang.PlaceAddress;
import NET.shudo.moba.reflect.*;
// ᥽å generatedPlaceFromFile() Τ
import java.io.FileInputStream;


/**
 * Thread internalizer
 *
 * @author SHUDOH Kazuyuki
 */
public
class ThreadReader extends ObjectReader {
  private static final String MOBA_PACKAGE =
	"NET.shudo.moba.lang.";
  private static final String STREAMABLE_THREAD =
	MOBA_PACKAGE + "MobaThread";
  private static Class STREAMABLE_THREAD_CLASS = null;

  private static MetaClass threadMC = null;
  private static MetaClass mobaThreadMC = null;

  static {
    try {
      STREAMABLE_THREAD_CLASS = Class.forName(STREAMABLE_THREAD);
    }
    catch (ClassNotFoundException e) {
      // not reached.
    }

    try {
      threadMC = MetaClass.get("java.lang.Thread");
      mobaThreadMC = MetaClass.get("NET.shudo.moba.lang.MobaThread");
    }
    catch (ClassNotFoundException e) {
      // not reached.
      System.err.println("java.lang.Thread or MobaThread is not found??");
      e.printStackTrace();  System.exit(1);
    }
  }


  /**
   * åɺ˽°٤ ThreadGroup
   */
  private ThreadGroup parentGroup = null;


  public ThreadReader(ThreadGroup g, InputStream in, ClassLoader loader)
	throws IOException {
    super(in, loader);
    this.parentGroup = g;
  }
  public ThreadReader(ThreadGroup g, InputStream in) throws IOException {
    this(g, in, null);
  }
  public ThreadReader(InputStream in, ClassLoader loader) throws IOException {
    this(null, in, loader);
  }
  public ThreadReader(InputStream in) throws IOException {
    this(null, in, null);
  }


  /**
   * 줿åɤǼ줿ե뤫顢
   * å롣
   */
  public static PlaceAddress generatedPlaceFromFile(String filename)
	throws IOException {
    PlaceAddress addr;
    ObjectReader oin = new ObjectReader(new FileInputStream(filename));

    oin.readUTF();	// skip 饹̾

/*
    try {
      addr = (PlaceAddress)oin.readObject();
    }
    catch (Exception e) {	// summarize some exception classes
      throw new IOException("getting thread generated place failed.");
    }
*/
    addr = new PlaceAddress();
    addr.readFromStream(oin);

    oin.close();
    return addr;
  }


  /**
   * subclassΥ饹Ф򤹤뤿hook
   * 󥹥󥹤֤
   *
   * @return true if the specified class is the target class.
   * @see NET.shudo.moba.io.ObjectReader#newInstanceOfSpecificClass
   */
  protected Object newInstanceOfSpecificClass(MetaClass mc) {
//	throws InstantiationException, IllegalAccessException {
    Object obj = mc.newInstance();

/*
    if (STREAMABLE_THREAD_CLASS.isAssignableFrom(mc.classClass())) {
//    if (mc.getName().equals(STREAMABLE_THREAD)) {	// 
    }
*/

    return obj;
  }

  /**
   * subclassΥ饹Ф򤹤뤿hook
   * ꥯ饹ɤȽꤹ롣
   *
   * @return true if the specified class is the target class.
   * @see NET.shudo.moba.io.ObjectReader#isSpecificClass
   */
  protected boolean isSpecificClass(Class clazz) {
    if (clazz.getName().equals(STREAMABLE_THREAD))
      return true;
    else
      return false;
  }

  /**
   * subclassΥ饹Ф򤹤뤿hook
   * Υ饹Ȥ supeclass ˤĤƤǤ롣
   *
   * @param obj instantiate ѤߤΥ֥
   * @return true if the specified class is the target class.
   * @see NET.shudo.moba.io.ObjectReader#readSpecificClass
   */
  protected boolean readSpecificClass(Class clazz, Object obj)
	throws StreamFormatException
		,IOException, ClassNotFoundException
		,InstantiationException, IllegalAccessException {
	/* override ObjectReader#readSpecificClass(Object) */
    if (clazz.getName().equals(STREAMABLE_THREAD)) {
      readThread((MobaThread)obj);
      return true;
    }
    else
      return false;
  }

  /**
   * subclassΥ饹Ф򤹤뤿hook<BR>
   * 󥹥󥹤ФԤ
   *
   * @return true if the specified class is the target class.
   * @see NET.shudo.moba.io.ObjectReader#postReadSpecificClass
   */
  protected boolean postReadSpecificClass(Object obj) {
    if (obj instanceof MobaThread) {
      ((MobaThread)obj).boot();
      return true;
    }
    else
      return false;
  }


  private MobaThread readThread(MobaThread thread)
	throws StreamFormatException
		,IOException, ClassNotFoundException
		,InstantiationException, IllegalAccessException {
    //  thread  instantiate Ѥ

    MetaField field = null;

    // internalized thread ǤȤɽ
    thread.isInternalized(true);


    try {
      // MobaThread#generatedPlace	åɤ
//      thread.generatedPlace((PlaceAddress)this.readObject());
      PlaceAddress genedPlace = new PlaceAddress();
      genedPlace.readFromStream(this);
      thread.generatedPlace(genedPlace);

      // MobaThread#hashValue
      field = mobaThreadMC.field("hashValue");
      field.setIntValue(thread, super.readInt());

      // MobaThread#generatedTime
      field = mobaThreadMC.field("generatedTime");
      field.setLongValue(thread, super.readLong());


      // (char[])name
      String thrName = null;
      field = threadMC.field("name");
      thrName = this.readUTF();
      field.setObjectValue(thread, (Object)thrName.toCharArray());

      // (int)priority
      field = threadMC.field("priority");
      field.setIntValue(thread, this.readInt());

      // (boolean)single_step
      field = threadMC.field("single_step");
      field.setBooleanValue(thread, this.readBoolean());

      // (boolean)daemon
      field = threadMC.field("daemon");
      field.setBooleanValue(thread, this.readBoolean());

      // (boolean)stillborn
      field = threadMC.field("stillborn");
      field.setBooleanValue(thread, this.readBoolean());

      // (Runnable)target
      field = threadMC.field("target");
      field.setObjectValue(thread, super.readObject());

      // (int)eetop  (ExecEnv *)
      MobaThread.readContextFrom(this, thread);


      // (ThreadGroup)group
      field = threadMC.field("group");
      {
	ThreadGroup g = null;
	if (parentGroup != null)	// °롼פλ꤬
	  g = parentGroup;
	else
	  g = Thread.currentThread().getThreadGroup();
	try {
	  MobaThread.setThreadGroup(thread, g);
	}
	catch (Exception e) {
	  // not reached.
	  System.out.println("In ThreadReader#readThread()");
	  System.out.println("  unexpected exception is caught:");
	  e.printStackTrace();
	}
      }
    }
    catch (FieldNotFoundException e) {
      // not reached.
      e.printStackTrace();  System.exit(1);
    }

    return thread;
  }
}
