/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.io;

import java.io.OutputStream;
import java.io.IOException;
import NET.shudo.moba.lang.MobaThread;
import NET.shudo.moba.lang.PlaceAddress;
import NET.shudo.moba.reflect.*;


/**
 * Thread externalizer
 *
 * @author SHUDOH Kazuyuki
 */
public
class ThreadWriter extends ObjectWriter {
  private static final String MOBA_PACKAGE =
	"NET.shudo.moba.lang.";
  private static final String STREAMABLE_THREAD =
	MOBA_PACKAGE + "MobaThread";

  private static MetaClass threadMC = null;
  private static MetaClass mobaThreadMC = null;

  static {
    try {
      threadMC = MetaClass.get("java.lang.Thread");
      mobaThreadMC = MetaClass.get("NET.shudo.moba.lang.MobaThread");
    }
    catch (ClassNotFoundException e) {
      // not reached.
      System.err.println("java.lang.Thread or MobaThread is not found??");
      e.printStackTrace();  System.exit(1);
    }
  }


  public ThreadWriter(OutputStream out) {
    super(out);
  }


  /**
   * subclassΥ饹Ф򤹤뤿hook
   *
   * @return true if the specified class is the target class.
   * @see NET.shudo.moba.io.ObjectWriter#isSpecificClass
   */
  protected boolean isSpecificClass(Class clazz) {
    if (clazz.getName().equals(STREAMABLE_THREAD))
      return true;
    else
      return false;
  }

  /**
   * subclassΥ饹Ф򤹤뤿hook
   * Υ饹Ȥ supeclass ˤĤƤǤ롣
   *
   * @return true if the specified class is the target class.
   * @see NET.shudo.moba.io.ObjectWriter#writeSpecificClass
   */
  protected boolean writeSpecificClass(Class clazz, Object obj)
	throws IOException {
	/* override ObjectWriter#writeSpecificClass(Object) */

    if (clazz.getName().equals(STREAMABLE_THREAD)) {
      MobaThread target = (MobaThread)obj;
      writeThread(target);

      if (target.isInternalized()) {
	// internalized åɤϤʾ write ƤϤʤ
	throw new StopMigrationException("Now stop writing. Don't worry.");
      }

      return true;
    }
    else {
      return false;
    }
  }


  /**
   * thread ν񤭽Ф
   */
  private void writeThread(MobaThread thr) throws IOException {
    Thread crntThread = Thread.currentThread();

    // internalized thread ǤϤʤȤɽ
    thr.isInternalized(false);

    if (thr.equals(crntThread)) {
      // ʬ (current thread)  write
      try {
	(new Thread(new SelfThreadWriter(this))).start();
      }
      catch (ClassNotWritableException e) {
	// not reached.
	System.err.println("an instance of MobaThread is not writable???.");
	e.printStackTrace();  System.exit(1);
      }

/*
      synchronized(selfWriter) {
	boolean notified = true;
	do {
	  try {
	    selfWriter.wait();
	  } catch (InterruptedException e) { notified = false; }
	} while (!notified);
      }
*/
      crntThread.suspend();

      return;
    }
    else {
      // ʬʳ write
      synchronized(thr) {
	thr.suspend();

	writeOtherThread(thr);

	thr.resume();
      }
    }
  }


  /**
   * current thread ʳ thread ν񤭽Ф
   */
  protected void writeOtherThread(MobaThread thr) throws IOException {
    MetaField field;

    try {
      // MobaThread#generatedPlace	åɤ
//      this.writeObject(thr.generatedPlace());
      thr.generatedPlace().writeToStream(this);

      // MobaThread#hashValue
      field = mobaThreadMC.field("hashValue");
      this.writeInt(field.getIntValue(thr));

      // MobaThread#generatedTime
      field = mobaThreadMC.field("generatedTime");
      this.writeLong(field.getLongValue(thr));


      // (char[])name
      field = threadMC.field("name");
      String thrName = new String((char[])field.getObjectValue(thr));
      this.writeUTF(thrName);

      // (int)priority
      field = threadMC.field("priority");
      this.writeInt(field.getIntValue(thr));

      // (boolean)single_step
      field = threadMC.field("single_step");
      this.writeBoolean(field.getBooleanValue(thr));

      // (boolean)daemon
      field = threadMC.field("daemon");
      this.writeBoolean(field.getBooleanValue(thr));

      // (boolean)stillborn
      field = threadMC.field("stillborn");
      this.writeBoolean(field.getBooleanValue(thr));

      // (Runnable)target
      field = threadMC.field("target");
      this.writeObject(field.getObjectValue(thr));

      // (int)eetop  (ExecEnv *)eetop
      thr.writeContextTo(this);
    }
    catch (FieldNotFoundException e) {
      // not reached.
      e.printStackTrace();  System.exit(1);
    }
  }
}


/**
 * current thread  externalize 뤿Υ饹
 *
 * @author SHUDOH Kazuyuki
 */
class SelfThreadWriter implements Runnable {
  private ThreadWriter out = null;
  private MobaThread target = null;
  public boolean success = true;

  public SelfThreadWriter(ThreadWriter out) throws ClassNotWritableException {
    this.out = out;

    Thread crntThread = Thread.currentThread();
    if (crntThread instanceof MobaThread) {
      this.target = (MobaThread)crntThread;
	//externalize оݤǤ current thread
    }
    else {
      throw new ClassNotWritableException(
	"current thread is not writable, must be instance of MobaThread.");
    }
  }

  public void run() {
    target.suspend();

    try {
      out.writeOtherThread(target);
    }
    catch (IOException e) {
      this.success = false;
    }

    target.resume();
/*
    synchronized(this) {
      this.notify();	// wake up target thread
    }
*/
  }
}
