/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include <stdio.h>	/* for NULL */
#include <netinet/in.h>	/* for hton[ls],ntoh[ls]() */

#include <native.h>	/* for native methods */
/*#include <sys_api.h>	/* for sys*() */
#include "sysmacros_md.h"	/* for sys*() */

#include "moba.h"


/*
 * Local Functions
 */
#define ARRAY_BLOCK_SIZE	100
static HArrayOfByte *array = NULL;

static void flushOutputStream(struct Hjava_io_OutputStream *out);


/*
 * Х (OutputStream) out ˽񤭽Ф
 *
 * returns: : 0, : -1
 */
extern int writeBytes(struct Hjava_io_OutputStream *out, char *data, int len) {
  HArrayOfByte array;

#ifdef WRITE_VERBOSE
  printf("writeBytes(out, 0x%x, %d) called.\n", (int)data, len);
  fflush(stdout);
#endif

  array.methods = mkatype(T_BYTE, len);
	/* mkatype() is defined in interpreter.h */
  array.obj = (ClassArrayOfByte *)data;

  return writeArrayOfByte(out, &array, 0, len);
	/* returns : 0, : -1 */
}


/*
 * Х (InputStream) in ɤ߹ࡣ
 * ǡΰѰդƤɬפϤʤ
 */
extern int readBytes(struct Hjava_io_InputStream *in, char *data, int len) {
  HArrayOfByte array;

  array.methods = mkatype(T_BYTE, len);
  array.obj = (ClassArrayOfByte *)data;

  return readArrayOfByte(in, &array, 0, len);
}


/*
 * ObjectWriter#writeObject(Object) ƤӽФ
 */
extern int writeObject(struct HNET_shudo_moba_io_ObjectWriter *out, JHandle *obj) {
  ExecEnv *ee = EE();
  static ClassClass *clazz = NULL;
  static struct methodblock *method = NULL;

  if (!clazz) {
    clazz = FindAndRegisterClass(ee, OBJWRITER_CLASS, FALSE, NULL);
	/* resolve is FALSE: superclass ϲ褷ʤ */
    if (!clazz) {
      /* not reached */
      printf("class " OBJWRITER_CLASS " is not found.\n");  fflush(stdout);
      return -1;
    }
  }

  if (!method) {
    method = find_method_in_class(ee, clazz, WRITEOBJ, WRITEOBJ_SIGNATURE);
    if (!method || exceptionOccurred(ee))  return -1;
  }
#ifdef DEBUG
  printf("writeObject():\n");
  printf("obj, out, clazz: 0x%x, 0x%x, 0x%x\n",
		(int)obj, (int)out, (int)clazz);
  if (obj) {
    printf("obj->methods: 0x%x\n", obj->methods);
    if (obj->methods)
      if (obj_flags(obj) == T_NORMAL_OBJECT)
	printf("class: %s\n", cbName(obj_classblock(obj)));
  }
  fflush(stdout);
#endif

  if (java_monitor) {
    struct methodblock *caller;
    long start_time, end_time;
    start_time = now();
    do_execute_java_method(ee, out,
	0, 0, method, 0, /* argument */ obj, NULL);
    end_time = now();
    caller = calleeByInvokeOpcode(
	ee->current_frame->current_method, ee->current_frame->lastpc);
    if (!caller)  caller = ee->current_frame->current_method;
    java_mon(caller, method, end_time-start_time);
  }
  else {
    do_execute_java_method(ee, out,
	0, 0, method, 0, /* argument */ obj, NULL);
  }

#ifdef DEBUG
  printf("writeObject() done.\n");
#endif
  if (exceptionOccurred(ee))
    return -1;

  return 0;
}


/*
 * ObjectReader#readObject() ƤӽФ
 */
extern JHandle *readObject(struct HNET_shudo_moba_io_ObjectReader *in) {
  ExecEnv *ee = EE();
  static ClassClass *clazz = NULL;
  static struct methodblock *method = NULL;
  JHandle *result = NULL;

  if (!clazz) {
    clazz = FindAndRegisterClass(ee, OBJREADER_CLASS, FALSE,
		unhand(in)->loader);
	/* resolve is FALSE: superclass ϲ褷ʤ */
    if (!clazz) {
      /* not reached */
      printf("class " OBJREADER_CLASS " is not found.\n");  fflush(stdout);
      return NULL;
    }
  }

  if (!method) {
    method = find_method_in_class(ee, clazz, READOBJ, READOBJ_SIGNATURE);
    if (!method || exceptionOccurred(ee))  return NULL;
  }

  if (java_monitor) {
    struct methodblock *caller;
    long start_time, end_time;
    start_time = now();
    result = (JHandle *)do_execute_java_method(ee, in,
		0, 0, method, 0, /* argument */ NULL);
    end_time = now();
    caller = calleeByInvokeOpcode(
	ee->current_frame->current_method, ee->current_frame->lastpc);
    if (!caller)  caller = ee->current_frame->current_method;
    java_mon(caller, method, end_time-start_time);
  }
  else {
    result = (JHandle *)do_execute_java_method(ee, in,
		0, 0, method, 0, /* argument */ NULL);
  }
#ifdef DEBUG
  if (exceptionOccurred(ee)) {
    printf("Exception occurred in ObjectReader#readObject().\n");
    fflush(stdout);

    exceptionDescribe(ee);
  }

  printf("in io.c, readObject() returns 0x%x\n", (int)result);
  fflush(stdout);
#endif
  return result;
}


/*
 * int64_t 񤭹ࡣ
 */
extern int writeInt64(Hjava_io_OutputStream *out, int64_t *p) {
#ifdef BIG_ENDIAN
  return writeBytes(out, (char *)p, 8);
#else
  int64_t v = hton64(*p);
  return writeBytes(out, (char *)&v, 8);
#endif
}


/*
 * int32_t 񤭹ࡣ
 */
extern int writeInt32(Hjava_io_OutputStream *out, int32_t *p) {
#ifdef BIG_ENDIAN
  return writeBytes(out, (char *)p, 4);
#else
  int32_t v;
  v = (int32_t)htonl((long)*p);
  return writeBytes(out, (char *)&v, 4);
#endif
}


/*
 * int16_t 񤭹ࡣ
 */
extern int writeInt16(Hjava_io_OutputStream *out, int16_t *p) {
#ifdef BIG_ENDIAN
  return writeBytes(out, (char *)p, 2);
#else
  int16_t v;
  v = (int16_t)htons((short)*p);
  return writeBytes(out, (char *)&v, 2);
#endif
}


/*
 * char 񤭹ࡣ
 */
extern int writeChar(Hjava_io_OutputStream *out, int c) {
  char ch = (char)c;
  return writeBytes(out, &ch, 1);
}


static void show64(char *x) {
  int i;
  char *p = (char *)x;

  printf("0x%02x", *p++);
  for (i = 1; i < 8; i++) {
    printf(", 0x%02x", *p++);
  }
  printf("\n");
  fflush(stdout);
}
/*
 * int64_t ɤ߹ࡣ
 */
extern int64_t readInt64(struct Hjava_io_InputStream *in) {
  int64_t v;
  if (readBytes(in, (char *)&v, 8) != 8)
    return -1;
#ifdef BIG_ENDIAN
  return v;
#else
  return (int64_t)ntoh64(v);
#endif
}


/*
 * int32_t ɤ߹ࡣ
 */
extern int32_t readInt32(struct Hjava_io_InputStream *in) {
  int32_t ret = 0;

  if (readBytes(in, (char *)&ret, 4) != 4)
    return -1;
#ifdef BIG_ENDIAN
  return ret;
#else
  return (int32_t)ntohl(*(long *)&ret);
#endif
}


/*
 * int16_t ɤ߹ࡣ
 */
extern int16_t readInt16(struct Hjava_io_InputStream *in) {
  int16_t ret = 0;

  if (readBytes(in, (char *)&ret, 2) != 2)
    return -1;
#ifdef BIG_ENDIAN
  return ret;
#else
  return (int16_t)ntohs(*(short *)&ret);
#endif
}


/*
 * char ɤ߹ࡣ
 */
extern char readChar(struct Hjava_io_InputStream *in) {
  char ret;

  if (readBytes(in, &ret, 1) != 1)
    return -1;
  return ret;
}


/*
 * ʸĹ, ʸ񤭽Ф
 *
 * returns: : 0, : -1
 */
extern int writeString(struct Hjava_io_OutputStream *out, char *str) {
  uint16_t len = (uint16_t)strlen(str);

  /* ʸĹ */
  if ((writeInt16(out, (int16_t *)&len)) < 0)  return -1;

  /* ʸ */
  if ((writeBytes(out, str, len)) < 0)  return -1;

  return 0;
}


/*
 * ʸĹ, ʸɤ߹ࡣ
 */
extern char *readStrdup(struct Hjava_io_InputStream *in) {
  uint16_t len;
  char *ptr;

  /* ʸĹ */
  len = (uint16_t)readInt16(in);

  if (!(ptr = (char *)sysMalloc((size_t)(len + 1)))) {
    SignalError(0, JAVAPKG "OutOfMemoryError", 0);
    return NULL;
  }

  /* ʸ */
  if (readBytes(in, ptr, (int)len) != len)
    return NULL;
  ptr[len] = '\0';	/* terminate */

  return ptr;
}


/*
 * class̾񤭽Ф
 */
extern int writeClassClass(struct Hjava_io_OutputStream *out, ClassClass *clazz) {
  char *className = cbName(clazz);

  if ((writeString(out, className)) < 0)  return -1;

  return 0;
}


/*
 * class̾ɤ߹ࡣ
 */
extern ClassClass *readClassClass(struct HNET_shudo_moba_io_ObjectReader *in) {
  ExecEnv *ee = EE();
  ClassClass *ret;
  char *className = readStrdup((Hjava_io_InputStream *)in);

  if (!className)  return NULL;

#ifdef DEBUG
  printf("class name: %s\n", className);
  fflush(stdout);
#endif
  ret = FindAndRegisterClass(ee, className, FALSE, unhand(in)->loader);
	/* resolve is FALSE: superclass ϲ褷ʤ */

  sysFree(className);

  return ret;
}


/*
 * struct methodblock 񤭽Ф
 */
extern int writeClassAndMethod(struct Hjava_io_OutputStream *out,
				struct methodblock *method) {
  ClassClass *clazz;
  char *name, *sig;

  clazz = method->fb.clazz;
  name = method->fb.name;
  sig = method->fb.signature;

#ifdef DEBUG
  printf("writeClassAndMethod():\n");
  printf(" class name: %s\n", cbName(clazz));
  printf(" method: %s %s\n", name, sig);
  fflush(stdout);
#endif
  if ((writeString(out, cbName(clazz))) < 0)  return -1;/* 饹̾ */
  if ((writeString(out, name)) < 0)  return -1;		/* ᥽å̾ */
  if ((writeString(out, sig)) < 0)  return -1;		/* signature */
#if 0
  printf("writeClassAndMethod() done.\n");
  fflush(stdout);
#endif

  return 0;
}


/*
 * struct methodblock ɤ߹ࡣ
 */
extern struct methodblock *readClassAndMethod(
	struct HNET_shudo_moba_io_ObjectReader* objIn) {
  ExecEnv *ee = EE();
  char *className, *name, *sig;
  ClassClass *clazz;
  struct methodblock *method = NULL;

  Hjava_io_InputStream *in = (Hjava_io_InputStream *)objIn;


  className = readStrdup(in);	/* 饹̾ */
  name = readStrdup(in);	/* ᥽å̾ */
  sig = readStrdup(in);		/* signature */
#ifdef DEBUG
  printf("class name: %s\n", className);
  printf("method: %s %s\n", name, sig);
  fflush(stdout);
#endif

  clazz = FindAndRegisterClass(ee, className, FALSE, unhand(objIn)->loader);
	/* resolve is FALSE: superclass ϲ褷ʤ */
  if (!clazz) {
    /* not reached */
    printf("class %s is not found.\n", className);  fflush(stdout);
    return 0;
  }

  method = find_method_in_class(ee, clazz, name, sig);
  if (exceptionOccurred(ee))  method = NULL;

  sysFree(className);
  sysFree(name);
  sysFree(sig);

  return method;
}


/*
 * stack_item 񤭽Ф
 * ֥Ȥ¨ͤϡ signature Ƚꤹ롣
 */
extern int writeStackItem(HNET_shudo_moba_io_ObjectWriter *out, stack_item *itemptr, char *sig) {
#ifdef DEBUG
  printf("writeStackItem((stack_item)0x%x, sig \"%s\") called.\n", *((int *)itemptr), sig);
  fflush(stdout);
#endif
  if (isSigObject(sig)) {
#ifdef DEBUG
    printf("stack item is handle.\n");
    fflush(stdout);
#endif
    if ((writeChar((Hjava_io_OutputStream *)out, STACK_ITEM_HANDLE)) < 0)
      return -1;
    if ((writeObject(out, /*item.h ǤϤʤ */ *((JHandle **)itemptr))) < 0)
      return -1;
	/* serialize */
  }
  else {
    if ((sig[0] == SIGNATURE_DOUBLE) || (sig[0] == SIGNATURE_LONG)) {
      Java8 tmp64;
      GET_INT64(tmp64, itemptr);
#ifdef DEBUG
      printf("stack item is 64bit immediate value: 0x%08x, 0x%08x\n",
		itemptr->i, (itemptr + 1)->i);
      printf(" substitute to: 0x%08x, 0x%08x\n",
		tmp64.x[0], tmp64.x[1]);
      fflush(stdout);
#endif
      if ((writeChar((Hjava_io_OutputStream *)out, STACK_ITEM_IMM64)) < 0)
	return -1;
      if ((writeInt64((Hjava_io_OutputStream *)out, (int64_t *)&tmp64)) < 0)
	return -1;
    }
    else {
#ifdef DEBUG
    printf("stack item is 32bit immediate value: %d\n", itemptr->i);
    fflush(stdout);
#endif
      if ((writeChar((Hjava_io_OutputStream *)out, STACK_ITEM_IMM32)) < 0)
	return -1;
      if ((writeInt32((Hjava_io_OutputStream *)out, (int32_t *)itemptr)) < 0)
	return -1;
    }
  }

  return 0;
}

/*
 * stack_item 񤭽Ф
 * ֥Ȥ¨ͤ signature 餺Ƚꤹ롣
 */
extern int writeStackItemAuto(struct HNET_shudo_moba_io_ObjectWriter *out, stack_item *itemptr) {
#ifdef DEBUG
  printf("writeStackItemAuto(0x%x) called.\n", *((int *)itemptr));
  fflush(stdout);
#endif
  if (isPtrHandle((void *)itemptr->h)) {
#ifdef DEBUG
    printf("stack item is handle.\n");
    fflush(stdout);
#endif
    if ((writeChar((Hjava_io_OutputStream *)out, STACK_ITEM_HANDLE)) < 0)
      return -1;
    if ((writeObject(out, itemptr->h)) < 0)  return -1;
	/* serialize */
  }
  else {
#ifdef DEBUG
    printf("stack item is immediate value.\n");
    fflush(stdout);
#endif
    if ((writeChar((Hjava_io_OutputStream *)out, STACK_ITEM_IMM32)) < 0)
      return -1;
    if ((writeInt32((Hjava_io_OutputStream *)out, ((int32_t *)itemptr))) < 0)
      return -1;
  }

#ifdef DEBUG
  printf("writeStackItemAuto() done.\n");
  fflush(stdout);
#endif
  return 0;
}

/*
 * stack_item ɤ߹ࡣ
 *
 * returns: item type (STACK_ITEM_*)
 */
extern int readStackItem(struct HNET_shudo_moba_io_ObjectReader *in, stack_item *dest) {
  char itemType = -1;
#ifdef DEBUG
    printf("readStackItem() called.\n");
    fflush(stdout);
#endif

  itemType = readChar((Hjava_io_InputStream *)in);
#ifdef DEBUG
    printf("item type is %d.\n", (int)itemType);
    fflush(stdout);
#endif

  switch ((int)itemType) {
  case STACK_ITEM_HANDLE:
#ifdef DEBUG
    printf("stack item is handle.\n");
    fflush(stdout);
#endif
    dest->h = (JHandle *)readObject(in);	/* de-serialize */
#ifdef DEBUG
    printf("readStackItem(): readObject() returns 0x%x\n", (int)dest);
    fflush(stdout);
#endif
    break;
  case STACK_ITEM_IMM32:
    dest->i = readInt32((Hjava_io_InputStream *)in);
#ifdef DEBUG
    printf("stack item is 32bit immediate value: %d\n", dest->i);
    fflush(stdout);
#endif
    break;
  case STACK_ITEM_IMM64:
    {
      int64_t val;
      Java8 tmp64;
      val = readInt64((Hjava_io_InputStream *)in);
      SET_INT64(tmp64, dest, val);	/* *((int64_t *)&dest) = val */
#ifdef DEBUG
      printf("stack item is 64bit immediate value: 0x%08x, 0x%08x\n",
		dest->i, (dest + 1)->i);
      printf(" substitute from: 0x%08x, 0x%08x\n",
		*((int32_t *)&val), *(((int32_t *)&val) + 1));
      fflush(stdout);
#endif
    }
    break;
  default:
#ifdef DEBUG
    printf("stream data is invalid...: 0x%X\n", (int)itemType & 0xff);
    fflush(stdout);
#endif
    SignalError(0, MOBA_PACKAGE "StreamFormatException",
		"STACK_ITEM_* is expected.");
    return;
    break;
  }
#ifdef DEBUG
  printf("readStackItem() done.\n");
  fflush(stdout);
#endif

  return (int)itemType;
}


/*
 * Local Functions
 */

#define WRITEBYTE_CLASSNAME	"java/io/OutputStream"
#define WRITEBYTE_SIGNATURE	"([BII)V"
#define WRITEBYTE_METHOD	"write"
#define READBYTE_CLASSNAME	"java/io/InputStream"
#define READBYTE_SIGNATURE	"([BII)I"
#define READBYTE_METHOD		"read"
#define FLUSH_CLASSNAME		"java/io/OutputStream"
#define FLUSH_SIGNATURE		"()V"
#define FLUSH_METHOD		"flush"


/*
 * OutputStream#write(byte[], int, int) ƤӽФ
 *
 * returns: : 0, 㳰ȯ: -1
 */
int writeArrayOfByte(struct Hjava_io_OutputStream *out,
	HArrayOfByte *data, int32_t off, int32_t len) {
  ExecEnv *ee = EE();
  static struct methodblock *method = NULL;

#ifdef WRITE_VERBOSE
  printf("writeArrayOfByte() called.\n");
  fflush(stdout);
#endif

  if (!method) {
    method = find_method_in_object(ee, (JHandle *)out,
				WRITEBYTE_METHOD, WRITEBYTE_SIGNATURE);
    if (!method) {
      /* not reached. */
      printf("method " WRITEBYTE_SIGNATURE " " WRITEBYTE_METHOD
	" is not found.\n");
      fflush(stdout);
      return -1;
    }
    if (exceptionOccurred(ee)) {
      exceptionDescribe(ee);
      fflush(stdout);
      return -1;
    }
  }
#ifdef WRITE_VERBOSE
  printf("%s#%s(0x%x, %d, %d)\n",
	cbName(method->fb.clazz), method->fb.name, (int)data, off, len);
  fflush(stdout);
#endif

  if (java_monitor) {
    struct methodblock *caller;
    long start_time, end_time;
    start_time = now();
    do_execute_java_method(ee, out,
	0, 0, method, 0, /* argument */ data, off, len, NULL);
    end_time = now();
    caller = calleeByInvokeOpcode(
	ee->current_frame->current_method, ee->current_frame->lastpc);
    if (!caller)  caller = ee->current_frame->current_method;
    java_mon(caller, method, end_time-start_time);
  }
  else {
    do_execute_java_method(ee, out,
	0, 0, method, 0, /* argument */ data, off, len, NULL);
  }

  if (exceptionOccurred(ee)) {
#ifdef DEBUG
    printf("io.c, writeArrayOfByte(): exception occurred:\n");
    fflush(stdout);
#endif
    exceptionDescribe(ee);
    fflush(stdout);
    return -1;
  }
  else
    return 0;
}


/*
 * InputStream#read(byte[]) ƤӽФ
 *
 * returns: ɤ᤿Хȿ, Ԥ -1
 */
int readArrayOfByte(struct Hjava_io_InputStream *in,
	HArrayOfByte *data, int32_t off, int32_t len) {
  ExecEnv *ee = EE();
  static struct methodblock *method = NULL;
  int32_t ret, readLen;

  char *body = ((ArrayOfByte *)unhand(data))->body;

#if 0
  printf("readArrayOfByte(in, 0x%x, %d, %d) called.\n", (int)data, off, len);
  fflush(stdout);
#endif

  if (!method) {
    method = find_method_in_object(ee, (JHandle *)in,
				READBYTE_METHOD, READBYTE_SIGNATURE);
    if (!method) {
      /* not reached. */
      printf("method " READBYTE_SIGNATURE " " READBYTE_METHOD
		" is not found.\n");
      fflush(stdout);
      return -1;
    }
    if (exceptionOccurred(ee)) {
      exceptionDescribe(ee);
      fflush(stdout);
      return -1;
    }
  }
#if 0
  printf("%s#%s(0x%x, %d, %d)\n",
	cbName(method->fb.clazz), method->fb.name, (int)data, off, len);
  fflush(stdout);
#endif

  readLen = 0;
  while (1) {
    if (java_monitor) {
      struct methodblock *caller;
      long start_time, end_time;
      start_time = now();
      ret = (int32_t)do_execute_java_method(ee, in,
		0, 0, method, 0, /* argument */ data, off, len, NULL);
      end_time = now();
      caller = calleeByInvokeOpcode(
	ee->current_frame->current_method, ee->current_frame->lastpc);
      if (!caller)  caller = ee->current_frame->current_method;
      java_mon(caller, method, end_time-start_time);
    }
    else {
      ret = (int32_t)do_execute_java_method(ee, in,
		0, 0, method, 0, /* argument */ data, off, len, NULL);
    }

#ifdef WRITE_VERBOSE
    {
      int i;
      printf("%s#%s((byte[%d])0x%x, %d, %d) returns %d.\n",
		cbName(method->fb.clazz), method->fb.name,
		obj_length(data), (int)data, off, len, ret);
      for (i = 0; i < ret; i++)
	printf(" %02x", (int)body[i] & 0xff);
      printf("\n");  fflush(stdout);
    }
#endif

    if (exceptionOccurred(ee)) {
#ifdef DEBUG
      printf("io.c, readArrayOfByte(): exception occurred:\n");
      fflush(stdout);
#endif
      exceptionDescribe(ee);
      fflush(stdout);
      return -1;
    }


    if (ret < 0)  break;	/* EOF */

    len -= ret;
    off += ret;
    readLen += ret;
    if (len <= 0)  break;
  }

  return readLen;
}


/*
 * OutputStream#flush() ƤӽФ
 */
static void flushOutputStream(struct Hjava_io_OutputStream *out) {
  ExecEnv *ee = EE();
  static struct methodblock *method = NULL;

  if (!method) {
    method = find_method_in_object(ee, (JHandle *)out,
				FLUSH_METHOD, FLUSH_SIGNATURE);
    if (!method) {
      /* not reached. */
      printf("method " FLUSH_SIGNATURE " " FLUSH_METHOD " is not found.\n");
      fflush(stdout);
      return;
    }
    if (exceptionOccurred(ee)) {
      exceptionDescribe(ee);
      fflush(stdout);
      return;
    }
  }

  if (java_monitor) {
    struct methodblock *caller;
    long start_time, end_time;
    start_time = now();
    do_execute_java_method(ee, out,
	FLUSH_METHOD, FLUSH_SIGNATURE, method, 0, /* argument */ NULL);
    end_time = now();
    caller = calleeByInvokeOpcode(
	ee->current_frame->current_method, ee->current_frame->lastpc);
    if (!caller)  caller = ee->current_frame->current_method;
    java_mon(caller, method, end_time-start_time);
  }
  else {
    do_execute_java_method(ee, out,
	FLUSH_METHOD, FLUSH_SIGNATURE, method, 0, /* argument */ NULL);
  }

  if (exceptionOccurred(ee)) {
    exceptionDescribe(ee);
    fflush(stdout);
  }
}
