/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/*
 * MOBA  native methods 
 */


#ifndef _MOBA_H_
#define _MOBA_H_


#include "mobanative.h"

/* for H<classname> */
#include "java_lang_ClassLoader.h"
#include "java_io_OutputStream.h"
#include "java_io_InputStream.h"
#include "NET_shudo_moba_io_ObjectWriter.h"
#include "NET_shudo_moba_io_ObjectReader.h"


/*
 * Definitions
 */

/*
 * Don't touch below stuffs
 */

/* endian */
/* GNU libc /usr/include/bytesex.h  */
#undef LITTLE_ENDIAN
#undef BIG_ENDIAN
#if defined(__i386__) || defined(__alpha__) \
	|| (defined(__mips__) && (defined(MIPSEL) || defined (__MIPSEL__)))
# define LITTLE_ENDIAN
#elif defined(__mc68000__) || defined (__sparc__) || defined (__PPC__) \
	|| (defined(__mips__) && (defined(MIPSEB) || defined (__MIPSEB__)))
# define BIG_ENDIAN
#else
# error architecture not supported by the Linux C library
#endif


#define MOBA_PACKAGE	"NET/shudo/moba/"
#define MOBA_REFLECT_PACKAGE	MOBA_PACKAGE "reflect/"
#define MOBA_IO_PACKAGE		MOBA_PACKAGE "io/"

#undef JDK_STYLE	/* JDK  Object Serialization Ȥݤ */

#ifdef JDK_STYLE
#define OBJECT_PACKAGE "java/io/"
#else
#define OBJECT_PACKAGE MOBA_IO_PACKAGE
#endif

#ifdef JDK_STYLE
#define OBJWRITER_CLASS	OBJECT_PACKAGE "ObjectOutputStream"
#define OBJREADER_CLASS	OBJECT_PACKAGE "ObjectInputStream"
#else
#define OBJWRITER_CLASS	OBJECT_PACKAGE "ObjectWriter"
#define OBJREADER_CLASS	OBJECT_PACKAGE "ObjectReader"
#endif

/* for util.c */
#define WRITEOBJ_SIGNATURE "(Ljava/lang/Object;)V"
#define READOBJ_SIGNATURE "()Ljava/lang/Object;"

#define WRITEOBJ "writeObject"
#define READOBJ "readObject"

/* for serialize.c */
#define BLOCK_IO_ARRAY

#define OWN_WRITEOBJ_SIGNATURE "(L" OBJWRITER_CLASS ";)V"
#define OWN_READOBJ_SIGNATURE "(L" OBJREADER_CLASS ";)V"

#define OWN_WRITEOBJ "writeObject"
#define OWN_READOBJ "readObject"


/* frame->constant_pool μ */
#define CP_CP		1
#define CP_METHOD	2	/* in case that frame->returnpc is 0 */


/* stack_item μ */
#define STACK_ITEM_HANDLE	2
#define STACK_ITEM_OBJECT	3
#define STACK_ITEM_ADDRESS	4
#define STACK_ITEM_IMM32	5
#define STACK_ITEM_IMM64	6


/* for meta*.c */
#if (JDK_VER < 110)	/* old fashion */
  #define HMetaClassToClassClass(/*(JHandle *)*/ handle) \
	((ClassClass *)unhand(unhand(handle)->thisClass))
#else
  #define HMetaClassToClassClass(/*(JHandle *)*/ handle) \
	((ClassClass *)(unhand(handle)->thisClass))
#endif /* JDK_VER < 110 */


/* for thread.c, serialize.c */
#define THROW_IOEXCEPTION() \
  {\
    ExecEnv* ee = EE();\
    struct Hjava_lang_Object *exc;\
    if (exceptionOccurred(ee)) {\
      exc = ee->exception.exc;\
    }\
    else {\
      exc = execute_java_constructor(ee,\
		"java.io.IOException", NULL, "()", NULL);\
    }\
    exceptionThrow(ee, exc);\
  }


/* #define COMPARE_CLASS_BY_NAME /* 饹(ClassClass *)Ӥ򥯥饹̾ */


#define min(A, B)	((A)<(B) ? (A) : (B))
#define max(A, B)	((A)>(B) ? (A) : (B))


/*
 * Global Functions
 */
/* thread.c */
JavaFrame *allocNewJavaFrame(struct methodblock *mb, ExecEnv *ee);
JavaFrame *allocNewJavaFrameWithNLocals(struct methodblock *mb, ExecEnv *ee,
					int nlocals);

/* signal_md.c */
extern int validateAddress(void *addr);

/* metaclass.c */
extern struct methodblock *find_own_rw_obj(ExecEnv *ee, ClassClass *cb, int flag_write);

/* constantpool.c */
extern char *cpTypeString(int type);
extern int searchUtf8(ClassClass *cb, char *utfstr);
extern int searchClazz(ClassClass *cb, ClassClass *targetCb);
extern int searchNameAndTypeByField(ClassClass *cb, struct fieldblock *fb);
extern int searchNameAndTypeByMethod(ClassClass *cb, struct methodblock *mb);
extern int searchNameAndType(ClassClass *cb, uint32_t id);

/* signature.c */
extern int isSigObject(char *sig);
extern int isSigArray(char *sig);

/* handle.c */
extern void registerGlobalHandle(HObject *obj);
extern void removeGlobalHandle(HObject *);

/* util.c */
extern struct methodblock *pc2method(unsigned char *pc);
extern ClassClass *cp2class(cp_item_type *cp);
extern ClassClass *FindAndRegisterClass(
	struct execenv *, char *name, bool_t resolve,
	struct Hjava_lang_ClassLoader *loader);
extern struct methodblock *find_method_in_class(ExecEnv *ee, ClassClass *cb,
					char *name, char *sig);
extern struct methodblock *find_method_in_object(ExecEnv *ee, JHandle *h,
					char *name, char *sig);
extern struct fieldblock *findFieldByID(ClassClass *cb, uint32_t ID);
extern struct methodblock *findMethodByID(ClassClass *cb, uint32_t ID);
extern struct fieldblock *findFieldInVM(uint32_t ID);
extern struct methodblock *findMethodInVM(uint32_t ID);

#define LocalVarTblLen(method)	((method)->localvar_table_length)
extern char *LocalVarName(struct methodblock *method, int index);
extern char *LocalVarSig(struct methodblock *method, int index);
#define LocalVarSlot(method, i)		(((method)->localvar_table)[i].slot)
#define LocalVarStartPC(method, i)	(((method)->localvar_table)[i].pc0)
#define LocalVarLengthPC(method, i)	(((method)->localvar_table)[i].length)
extern int isLocalVarAlive(struct methodblock *, int, uint32_t pcoff);

extern unsigned char *nextPC(unsigned char *pc);
extern void dumpOpcodeLength(FILE *out);

extern int isPtrHandle(void *p);

extern HArrayOfByte *makeArrayOfByte(char *str, int len);
extern void outUnicode(FILE *out, unicode *str);
extern void outPC(FILE *out, unsigned char *pc, struct methodblock *method);
extern void outMethod(FILE *out, struct methodblock *method);

extern uint32_t pc2offset(struct methodblock *method, unsigned char *pc);
extern unsigned char *offset2pc(struct methodblock *method, uint32_t offset);
extern int16_t sp2offset(JavaFrame *frame, stack_item *sp);
extern stack_item *offset2sp(JavaFrame *frame, int16_t offset);
extern uint16_t cp2offset(ClassClass *clazz, cp_item_type *cp_item);
extern cp_item_type *offset2cp(ClassClass *clazz, uint16_t offset);

extern struct methodblock *calleeByInvokeOpcode(
			struct methodblock *mb, unsigned char *lastpc);

extern int signature2typecode(char *sig);

#ifdef BIG_ENDIAN
#define ntoh64(X) (X)
#define hton64(X) (X)
#else
extern uint64_t hton64(uint64_t x);
extern uint64_t ntoh64(uint64_t x);
#endif

/* io.c */
int writeArrayOfByte(Hjava_io_OutputStream *out,
	HArrayOfByte *data, int32_t off, int32_t len);
	/* call OutputStream.write(byte[], int, int) */
int readArrayOfByte(Hjava_io_InputStream *in,
	HArrayOfByte *data, int32_t off, int32_t len);
	/* call InputStream.read(byte[], int, int) */

extern int writeBytes(Hjava_io_OutputStream *out, char *data, int len);
extern int readBytes(Hjava_io_InputStream *in, char *data, int len);
extern int writeObject(HNET_shudo_moba_io_ObjectWriter *out, JHandle *obj);
extern JHandle *readObject(HNET_shudo_moba_io_ObjectReader *in);

extern int writeInt64(Hjava_io_OutputStream *out, int64_t *p);
extern int writeInt32(Hjava_io_OutputStream *out, int32_t *p);
extern int writeInt16(Hjava_io_OutputStream *out, int16_t *p);
extern int writeChar(Hjava_io_OutputStream *out, int c);
extern int64_t readInt64(Hjava_io_InputStream *in);
extern int32_t readInt32(Hjava_io_InputStream *in);
extern int16_t readInt16(Hjava_io_InputStream *in);
extern char readChar(Hjava_io_InputStream *in);

extern int writeString(Hjava_io_OutputStream *out, char *);
extern char *readStrdup(Hjava_io_InputStream *in);

extern int writeClassClass(Hjava_io_OutputStream *out, ClassClass *cb);
extern ClassClass *readClassClass(HNET_shudo_moba_io_ObjectReader *in);
extern int writeClassAndMethod(
	Hjava_io_OutputStream *out, struct methodblock *method);
extern struct methodblock *readClassAndMethod(
	HNET_shudo_moba_io_ObjectReader *in);
extern int writeStackItem(
	HNET_shudo_moba_io_ObjectWriter *out, stack_item *itemptr, char *sig);
extern int writeStackItemAuto(HNET_shudo_moba_io_ObjectWriter *out, stack_item *itemptr);
extern int readStackItem(HNET_shudo_moba_io_ObjectReader *in, stack_item *dest);


#endif /* _MOBA_H_ */
