/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

/* native methods for moba.reflect.Signature */

#include <string.h>	/* strlen() */

#include "NET_shudo_moba_reflect_Signature.h"

#include "moba.h"


/*
 * Local Functions
 */
static int isSigString(char *sig);
static int isSigBaseType(char *sig);
static int isSigObjectArray(char *sig);
static int isSigBaseTypeArray(char *sig);

static int is_same_type(struct Hjava_lang_String *sig, int sig_head);


extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isObject(struct HNET_shudo_moba_reflect_Signature *_this ,struct Hjava_lang_String *str) {
  char *sig = makeCString(str);
  return isSigObject(sig);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isString(struct HNET_shudo_moba_reflect_Signature *_this ,struct Hjava_lang_String *str) {
  char *sig = makeCString(str);
  return isSigString(sig);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isBaseType(struct HNET_shudo_moba_reflect_Signature *_this ,struct Hjava_lang_String *str) {
  char *sig = makeCString(str);
  return isSigBaseType(sig);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isArray(struct HNET_shudo_moba_reflect_Signature *_this ,struct Hjava_lang_String *str) {
  char *sig = makeCString(str);
  return isSigArray(sig);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isObjectArray(struct HNET_shudo_moba_reflect_Signature *_this ,struct Hjava_lang_String *str) {
  char *sig = makeCString(str);
  return isSigObjectArray(sig);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isBaseTypeArray(struct HNET_shudo_moba_reflect_Signature *_this ,struct Hjava_lang_String *str) {
  char *sig = makeCString(str);
  return isSigBaseTypeArray(sig);
}


/*
 * ΥեɤηǤ뤫ɤĴ٤롣
 * Any, byte, char, double, enum, float, int, long, short, void, boolean
 */

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isAny(struct HNET_shudo_moba_reflect_Signature *_this, struct Hjava_lang_String *sig) {
  return is_same_type(sig, SIGNATURE_ANY);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isByte(struct HNET_shudo_moba_reflect_Signature *_this, struct Hjava_lang_String *sig) {
  return is_same_type(sig, SIGNATURE_BYTE);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isChar(struct HNET_shudo_moba_reflect_Signature *_this, struct Hjava_lang_String *sig) {
  return is_same_type(sig, SIGNATURE_CHAR);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isDouble(struct HNET_shudo_moba_reflect_Signature *_this, struct Hjava_lang_String *sig) {
  return is_same_type(sig, SIGNATURE_DOUBLE);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isEnum(struct HNET_shudo_moba_reflect_Signature *_this, struct Hjava_lang_String *sig) {
  return is_same_type(sig, SIGNATURE_ENUM);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isFloat(struct HNET_shudo_moba_reflect_Signature *_this, struct Hjava_lang_String *sig) {
  return is_same_type(sig, SIGNATURE_FLOAT);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isInt(struct HNET_shudo_moba_reflect_Signature *_this, struct Hjava_lang_String *sig) {
  return is_same_type(sig, SIGNATURE_INT);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isLong(struct HNET_shudo_moba_reflect_Signature *_this, struct Hjava_lang_String *sig) {
  return is_same_type(sig, SIGNATURE_LONG);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isShort(struct HNET_shudo_moba_reflect_Signature *_this, struct Hjava_lang_String *sig) {
  return is_same_type(sig, SIGNATURE_SHORT);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isVoid(struct HNET_shudo_moba_reflect_Signature *_this, struct Hjava_lang_String *sig) {
  return is_same_type(sig, SIGNATURE_VOID);
}

extern /*boolean*/ int32_t NET_shudo_moba_reflect_Signature_isBoolean(struct HNET_shudo_moba_reflect_Signature *_this, struct Hjava_lang_String *sig) {
  return is_same_type(sig, SIGNATURE_BOOLEAN);
}


/*
 * Local Functions
 */

/*
 * true if it's object or array
 */
int isSigObject(char *sig) {
  return ((sig[0] == SIGNATURE_CLASS) || (sig[0] == SIGNATURE_ARRAY));
}

/*
 * true if it's array
 */
int isSigArray(char *sig) {
  return (sig[0] == SIGNATURE_ARRAY);
}

/*
 * object of class java.lang.String
 */
static int isSigString(char *sig) {
  static char *sigString = "Ljava/lang/String;";
  int cmpLen = strlen(sigString);;

  return !(strncmp(sig, sigString, cmpLen));
}

static int isSigBaseType(char *sig) {
  return ((sig[0] != SIGNATURE_CLASS) && (sig[0] != SIGNATURE_ARRAY));
}

static int isSigObjectArray(char *sig) {
  int len = strlen(sig);

  if (sig[0] == SIGNATURE_ARRAY) {
    if (len >= 4) {
      return (isSigObject(sig + 1));
    }
  }
  return 0;	/* false */
}

static int isSigBaseTypeArray(char *sig) {
  int len = strlen(sig);

  if (sig[0] == SIGNATURE_ARRAY) {
    if (len >= 2) {
      return (isSigBaseType(sig + 1));
    }
  }
  return 0;	/* false */
}


static int is_same_type(struct Hjava_lang_String *sig, int sig_head) {
  return (makeCString(sig)[0] == sig_head);
}
