/**
  Sample code for thread migration
  in package NET.shudo.moba.

  Copy this file to the directory which is in CLASSPATH and compile.
    % javac MigrationThread.java
  Make sure that this class (MigrationThread) is found in your CLASSPATH
  at the source host.
  Type the following command at the source and destination host
    % (somewhere)/moba/bin/mobaplace &
  to invoke MOBA place which receive threads from the network.
  Now execute at the source host.
    % java MigrationThread <destination_host_name>

  @author SHUDOH Kazuyuki
 */

import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.PrintStream;
import java.io.IOException;

import NET.shudo.moba.lang.MobaThread;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Date;
import java.rmi.*;


class MigrationThread {
  private void start(String[] argv) {
    // åɤεư
    Thread t = null;
    if (argv.length > 0)
      t = new AThreadMigrant(argv[0]);
    else
      t = new AThreadMigrant();
    t.start();

    // ưåɤνλԤ
    boolean joined = true;
    do {
      try {
	t.join();
      } catch (InterruptedException e) { joined = false; }
    } while (!joined);
  }

  public static void main(String[] argv) {
    new MigrationThread().start(argv);
  }
}


/**
 * ư륹å
 */
class AThreadMigrant extends MobaThread {
  private static final String DEFAULT_HOST = "host.defaultdomain.com";
  private static final String THREAD_FNAME = "output";

  private String host = null;

  private InetAddress srcAddr = null;

  public AThreadMigrant() {
    this(DEFAULT_HOST);
  }
  public AThreadMigrant(String host) {
    this.host = host;
  }


  public void run() {
    long time0, time1;
    long remoteObjRet = -1;

    System.out.println("Now externalize myself(Thread#currentThread()).");

    // 򵭲
    try {
      srcAddr = InetAddress.getLocalHost();
    }
    catch (UnknownHostException e) {
      e.printStackTrace();
      Thread.currentThread().stop();
    }

    time0 = (new Date()).getTime();

    // ư
    try {
      MobaThread.goTo(this.host);
    }
    catch (IOException e) {
      e.printStackTrace();
      Thread.currentThread().stop();
    }

System.out.println("time0: " + time0);
System.out.flush();

    // ׻
    try {
System.out.println("0");
System.out.flush();
      bench.rmi.CalcServer remoteObj =
	(bench.rmi.CalcServer)Naming.lookup("//" + this.host + "/CalcServer");
System.out.println("1");
System.out.flush();
      remoteObjRet = remoteObj.arrayTransfer(new double[1]);
System.out.println("remote object returns: " + remoteObjRet);
System.out.flush();
    }
    catch (Exception e) {
      e.printStackTrace();
      Thread.currentThread().stop();
    }

    // ư
    try {
      MobaThread.goTo(srcAddr);
    }
    catch (IOException e) {
      e.printStackTrace();
      Thread.currentThread().stop();
    }

    time1 = (new Date()).getTime();

    System.out.println("I've returned !");
    System.out.println("remote object returns: " + remoteObjRet);
    System.out.println("time: " + (time1 - time0) + " (ms)");
  }
}
